package wsx2.dojapcmsound;

class G721encoder {
	/*
	 * The following is the definition of the state structure
	 * used by the G.721/G.723 encoder and decoder to preserve their internal
	 * state between successive calls.  The meanings of the majority
	 * of the state structure fields are explained in detail in the
	 * CCITT Recommendation G.721.  The field names are essentially indentical
	 * to variable names in the bit level description of the coding algorithm
	 * included in this Recommendation.
	 */
	class g72x_state {
		long yl;	/* Locked or steady state step size multiplier. */
		int yu;	/* Unlocked or non-steady state step size multiplier. */
		int dms;	/* Short term energy estimate. */
		int dml;	/* Long term energy estimate. */
		int ap;	/* Linear weighting coefficient of 'yl' and 'yu'. */

		int a[] = new int[2];	/* Coefficients of pole portion of prediction filter. */
		int b[] = new int[6];	/* Coefficients of zero portion of prediction filter. */
		int pk[] = new int[2];	/*
		 * Signs of previous two samples of a partially
		 * reconstructed signal.
		 */
		short dq[] = new short[6];	/*
		 * Previous 6 samples of the quantized difference
		 * signal represented in an internal floating point
		 * format.
		 */
		short sr[] = new short[2];	/*
		 * Previous 2 samples of the quantized difference
		 * signal represented in an internal floating point
		 * format.
		 */
	}

	private int power2[] = {1, 2, 4, 8, 0x10, 0x20, 0x40, 0x80,
			0x100, 0x200, 0x400, 0x800, 0x1000, 0x2000, 0x4000};

	private int qtab_721[] = {-124, 80, 178, 246, 300, 349, 400};
	/*
	 * Maps G.721 code word to reconstructed scale factor normalized log
	 * magnitude values.
	 */
	// -1, 1, 2, 3, 4, 6, 7,10
	private int	_dqlntab[] = {-2048, 4, 135, 213, 273, 323, 373, 425,
			425, 373, 323, 273, 213, 135, 4, -2048};

	/* Maps G.721 code word to log of scale factor multiplier. */
//	private int	_witab[] = {-12, 18, 41, 64, 112, 198, 355, 1122,
//			1122, 355, 198, 112, 64, 41, 18, -12};
	
	private int	_witab[] = {-384, 576, 1312, 2048, 3584, 6336, 11360, 35904, 
			35904, 11360, 6336, 3584, 2048, 1312, 576, -384 };
	/*
	 * Maps G.721 code words to a set of values whose long and int
	 * term averages are computed and then compared to give an indication
	 * how stationary (steady state) the signal is.
	 */
	private int	_fitab[] = {0, 0, 0, 0x200, 0x200, 0x200, 0x600, 0xE00,
			0xE00, 0x600, 0x200, 0x200, 0x200, 0, 0, 0};

	/*
	 * predictor_zero()
	 *
	 * computes the estimated signal from 6-zero predictor.
	 *
	 */
	private int
	predictor_zero(g72x_state state_ptr) {
		int		i;
		int		sezi;

		sezi = fmult(state_ptr.b[0] >> 2, state_ptr.dq[0]);
		for (i = 1; i < 6; i++)			/* ACCUM */
			sezi += fmult(state_ptr.b[i] >> 2, state_ptr.dq[i]);
		return (sezi);
	}

	/*
	 * predictor_pole()
	 *
	 * computes the estimated signal from 2-pole predictor.
	 *
	 */
	private int predictor_pole(g72x_state state_ptr) {
		return (fmult(state_ptr.a[1] >> 2, state_ptr.sr[1]) +
				fmult(state_ptr.a[0] >> 2, state_ptr.sr[0]));
	}



	/*
	 * fmult()
	 *
	 * returns the integer product of the 14-bit integer "an" and
	 * "floating point" representation (4-bit exponent, 6-bit mantessa) "srn".
	 */
	private int fmult(int an, int srn) {
		int		anmag, anexp, anmant;
		int		wanexp, wanmant;
		int		retval;
	
		anmag = (an > 0) ? an : ((-an) & 0x1FFF);
		anexp = quan(anmag, power2) - 6;
		anmant = (anmag == 0) ? 32
				:(anexp >= 0) ? anmag >> anexp
				: anmag << -anexp;
		wanexp = anexp + ((srn >> 6) & 0xF) - 13;
	
		wanmant = (anmant * (srn & 077) + 0x30) >> 4;
		retval = (wanexp >= 0) ? ((wanmant << wanexp) & 0x7FFF) :
			(wanmant >> -wanexp);
	
		return (((an ^ srn) < 0) ? -retval : retval);
	}

	/*
	 * step_size()
	 *
	 * computes the quantization step size of the adaptive quantizer.
	 *
	 */
	private int step_size(g72x_state state_ptr) {
		int		y;
		int		dif;
		int		al;
	
		if (state_ptr.ap >= 256)
			return (state_ptr.yu);
		else {
			y = (int)(state_ptr.yl >> 6);
			dif = state_ptr.yu - y;
			al = state_ptr.ap >> 2;
			if (dif > 0)
				y += (dif * al) >> 6;
				else if (dif < 0)
					y += (dif * al + 0x3F) >> 6;
					return (y);
		}
	}

	/*
	 * quan()
	 *
	 * quantizes the input val against the table of size short integers.
	 * It returns i if table[i - 1] <= val < table[i].
	 *
	 * Using linear search for simple coding.
	 */
	private int quan(int val, int table[] ) {
		int		i;
		int		size;
		
		size = table.length;

		for (i = 0; i < size; i++)
			if (val < table[i])
				break;
		return (i);
	}

	/*
	 * quantize()
	 *
	 * Given a raw sample, 'd', of the difference signal and a
	 * quantization step size scale factor, 'y', this routine returns the
	 * ADPCM codeword to which that sample gets quantized.  The step
	 * size scale factor division operation is done in the log base 2 domain
	 * as a subtraction.
	 */
	private int quantize(
			int		d,	/* Raw difference signal sample */
			int		y	/* Step size multiplier */
	) 
	{

		int		dqm;	/* Magnitude of 'd' */
		int		exp;	/* Integer part of base 2 log of 'd' */
		int		mant;	/* Fractional part of base 2 log */
		int		dl;	/* Log of magnitude of 'd' */
		int		dln;	/* Step size scale factor normalized log */
		int		i;

		/*
		 * LOG
		 *
		 * Compute base 2 log of 'd', and store in 'dl'.
		 */
		dqm = Math.abs(d);
		exp = quan(dqm >> 1, power2);
		mant =((dqm << 7) >> exp) & 0x7F;	/* Fractional portion. */
		dl = (exp << 7) + mant;

		/*
		 * SUBTB
		 *
		 * "Divide" by step size multiplier.
		 */
		dln = dl - (y >> 2);

		/*
		 * QUAN
		 *
		 * Obtain codword i for 'd'.
		 */
		i = quan(dln, qtab_721);
		if (d < 0)			/* take 1's complement of i */
			return ((qtab_721.length << 1) + 1 - i);
		else if (i == 0)		/* take 1's complement of 0 */
			return ((qtab_721.length << 1) + 1); /* new in 1988 */
		else
			return (i);
	}
	
	/*
	 * reconstruct()
	 *
	 * Returns reconstructed difference signal 'dq' obtained from
	 * codeword 'i' and quantization step size scale factor 'y'.
	 * Multiplication is performed in log base 2 domain as addition.
	 */
	private int reconstruct(
			int		sign,	/* 0 for non-negative value */
			int		dqln,	/* G.72x codeword */
			int		y)	/* Step size multiplier */
	{
		int		dql;	/* Log of 'dq' magnitude */
		int		dex;	/* Integer part of log */
		int		dqt;
		int		dq;	/* Reconstructed difference signal sample */
	
		dql = dqln + (y >> 2);	/* ADDA */
	
		if (dql < 0) {
			return ((0 != sign) ? -0x8000 : 0);
		} else {		/* ANTILOG */
			dex = (dql >> 7) & 15;
			dqt = 128 + (dql & 127);
			dq = (dqt << 7) >> (14 - dex);
			return ((0 != sign) ? (dq - 0x8000) : dq);
		}
	}

	/*
	 * update()
	 *
	 * updates the state variables for each output code
	 */
	private void update(
			int		y,		/* quantizer step size */
			int		code,	/* ADPCM code */
			int		dq,		/* quantized prediction difference */
			int		sr,		/* reconstructed signal */
			int		dqsez,		/* difference from 2-pole predictor */
			g72x_state state_ptr)	/* coder state pointer */
	{
		int		cnt;
		int		mag, exp;	/* Adaptive predictor, FLOAT A */
		int		a2p = 0;		/* LIMC */
		int		a1ul;		/* UPA1 */
		int		pks1;	/* UPA2 */
		int		fa1;
		int		pk0;

//		int		wi = _witab[code] << 5; /* scale factor multiplier */
		int		wi = _witab[code]; /* scale factor multiplier */
		int		fi = _fitab[code];		/* for long/short term energies */

		pk0 = (dqsez < 0) ? 1 : 0;	/* needed in updating predictor poles */
		mag = dq & 0x7FFF;		/* prediction difference magnitude */

		/*
		 * Quantizer scale factor adaptation.
		 */

		/* FUNCTW & FILTD & DELAY */
		/* update non-steady state step size multiplier */
		state_ptr.yu = y + ((wi - y) >> 5);

		/* LIMB */
		if (state_ptr.yu < 544)	/* 544 <= yu <= 5120 */
			state_ptr.yu = 544;
		else if (state_ptr.yu > 5120)
			state_ptr.yu = 5120;

		/* FILTE & DELAY */
		/* update steady state step size multiplier */
		state_ptr.yl += state_ptr.yu + ((-state_ptr.yl) >> 6);


		/*
		 * Adaptive predictor coefficients.
		 */
		/* update a's and b's */
		pks1 = pk0 ^ state_ptr.pk[0];		/* UPA2 */

		/* update predictor pole a[1] */
		a2p = state_ptr.a[1] - (state_ptr.a[1] >> 7);
		if (dqsez != 0) {
			fa1 = (0 != pks1) ? state_ptr.a[0] : -state_ptr.a[0];
			if (fa1 < -8191)	/* a2p = function of fa1 */
				a2p -= 0x100;
			else if (fa1 > 8191)
				a2p += 0xFF;
			else
				a2p += fa1 >> 5;

				if (0 != (pk0 ^ state_ptr.pk[1]))
					/* LIMC */
					if (a2p <= -12160)
						a2p = -12288;
					else if (a2p >= 12416)
						a2p = 12288;
					else
						a2p -= 0x80;
				else if (a2p <= -12416)
					a2p = -12288;
				else if (a2p >= 12160)
					a2p = 12288;
				else
					a2p += 0x80;
		}

		/* TRIGB & DELAY */
		state_ptr.a[1] = a2p;

		/* UPA1 */
		/* update predictor pole a[0] */
		state_ptr.a[0] -= state_ptr.a[0] >> 8;
		if (dqsez != 0)
			if (pks1 == 0)
				state_ptr.a[0] += 192;
			else
				state_ptr.a[0] -= 192;

		/* LIMD */
		a1ul = 15360 - a2p;
		if (state_ptr.a[0] < -a1ul)
			state_ptr.a[0] = -a1ul;
		else if (state_ptr.a[0] > a1ul)
			state_ptr.a[0] = a1ul;

		/* UPB : update predictor zeros b[6] */
		for (cnt = 0; cnt < 6; cnt++) {
			state_ptr.b[cnt] -= state_ptr.b[cnt] >> 8;

			if (0 != (dq & 0x7FFF)) {			/* XOR */
				if ((dq ^ state_ptr.dq[cnt]) >= 0)
					state_ptr.b[cnt] += 128;
				else
					state_ptr.b[cnt] -= 128;
			}
		}



		for (cnt = 5; cnt > 0; cnt--)
			state_ptr.dq[cnt] = state_ptr.dq[cnt-1];
		/* FLOAT A : convert dq[0] to 4-bit exp, 6-bit mantissa f.p. */
		if (mag == 0) {
			state_ptr.dq[0] = (short)((dq >= 0) ? 0x20 : 0xFC20);
		} else {
			exp = quan(mag, power2);
			state_ptr.dq[0] = (short)(
					(dq >= 0) ?
							(exp << 6) + ((mag << 6) >> exp) :
								(exp << 6) + ((mag << 6) >> exp) - 0x400
			);
		}

		state_ptr.sr[1] = state_ptr.sr[0];
		/* FLOAT B : convert sr to 4-bit exp., 6-bit mantissa f.p. */
		if (sr == 0) {
			state_ptr.sr[0] = 0x20;
		} else if (sr > 0) {
			exp = quan(sr, power2);
			state_ptr.sr[0] = (short)((exp << 6) + ((sr << 6) >> exp));
		} else if (sr > -32768) {
			mag = -sr;
			exp = quan(mag, power2);
			state_ptr.sr[0] =  (short)((exp << 6) + ((mag << 6) >> exp) - 0x400);
		} else
			state_ptr.sr[0] = (short)0xFC20;

		/* DELAY A */
		state_ptr.pk[1] = state_ptr.pk[0];
		state_ptr.pk[0] = pk0;


		/*
		 * Adaptation speed control.
		 */
		state_ptr.dms += (fi - state_ptr.dms) >> 5;		/* FILTA */
		state_ptr.dml += (((fi << 2) - state_ptr.dml) >> 7);	/* FILTB */

		if (y < 1536)					/* SUBTC */
			state_ptr.ap += (0x200 - state_ptr.ap) >> 4;
		else if (Math.abs((state_ptr.dms << 2) - state_ptr.dml) >=
			(state_ptr.dml >> 3))
			state_ptr.ap += (0x200 - state_ptr.ap) >> 4;
		else
			state_ptr.ap += (-state_ptr.ap) >> 4;
	}

	/*
	 * g721_encoder()
	 *
	 * Encodes the input vale of linear PCM, A-law or u-law data sl and returns
	 * the resulting code. -1 is returned for unknown input coding value.
	 */
	private int g721_encoder(int sl, g72x_state state_ptr) {
		int		sezi, se, sez;		/* ACCUM */
		int		d;			/* SUBTA */
		int		sr;			/* ADDB */
		int		y;			/* MIX */
		int		dqsez;			/* ADDC */
		int		dq;
		int i;
	
		sl >>= 2;			/* 14-bit dynamic range */
	
	
		sezi = predictor_zero(state_ptr);	//computes the estimated signal from 6-zero predictor.
		sez = sezi >> 1;
		se = (sezi + predictor_pole(state_ptr)) >> 1;	/* 14bitPCM \l */
	
		d = sl - se;				/* 14bitPCM \l */
	
		/* quantize the prediction difference */
		y = step_size(state_ptr);		/* quantizer step size */
		i = quantize(d, y);	/* i = ADPCM code */
	
		dq = reconstruct(i & 8, _dqlntab[i], y);	/* quantized est diff */
	
		sr = (dq < 0) ? se - (dq & 0x3FFF) : se + dq;	/* reconst. signal */
	
		dqsez = (sr + sez - se);			/* pole prediction diff. */
	
		update(y, i, dq, sr, dqsez, state_ptr);
	
		return (i);
	}

	/*
	 * g72x_init_state()
	 *
	 * This routine initializes and/or resets the g72x_state structure
	 * pointed to by 'state_ptr'.
	 * All the initial state values are specified in the CCITT G.721 document.
	 */
	private void g72x_init_state(g72x_state state_ptr) {
		int		cnta;
	
		state_ptr.yl = 34816;
		state_ptr.yu = 544;
		state_ptr.dms = 0;
		state_ptr.dml = 0;
		state_ptr.ap = 0;
		for (cnta = 0; cnta < 2; cnta++) {
			state_ptr.a[cnta] = 0;
			state_ptr.pk[cnta] = 0;
			state_ptr.sr[cnta] = 32;
		}
		for (cnta = 0; cnta < 6; cnta++) {
			state_ptr.b[cnta] = 0;
			state_ptr.dq[cnta] = 32;
		}
	}






	private g72x_state stat;

	public G721encoder() {
		stat = new g72x_state();
		g72x_init_state(stat);
	}

	public int encode(int sample) {
		return g721_encoder(sample, stat);

	}

	public void initState() {
		g72x_init_state(stat);
	}

}
